import geopandas as gpd
import pandas as pd


def add_zone_ids(catalog: pd.DataFrame,
                 zones: gpd.GeoDataFrame,
                 zone_column_name: str) -> pd.DataFrame:
    """
    Adds a column zone_column_name to the catalog,
    based on in which zone the event is located.
    Note that some events may not get a zone assigned.

    Args:
        catalog: The catalog with required columns: latitude, longitude
        zones: GeoDataFrame with the zones, required column `zone_column_name`

    Returns:
        the catalog with the `zone_column_name` column added
    """
    if zone_column_name not in zones.columns:
        raise ValueError(
            f'{zone_column_name} not found in the tectonic source shapefile')
    if 'latitude' not in catalog.columns:
        raise ValueError('latitude not found in the catalog')
    if 'longitude' not in catalog.columns:
        raise ValueError('longitude not found in the catalog')
    zones = zones.set_index(zone_column_name, drop=True)
    geometry = gpd.points_from_xy(
        catalog['longitude'], catalog['latitude'], crs="EPSG:4326")
    if zone_column_name in catalog.columns:
        catalog = catalog.drop(columns=[zone_column_name])
    geo_catalog = gpd.GeoDataFrame(catalog, geometry=geometry)
    geo_catalog = geo_catalog.sjoin(zones[['geometry']],
                                    how='left', predicate='within')
    return geo_catalog
