import os

import geopandas as gpd
import numpy as np
import pandas as pd

from psha.common import add_zone_ids


def _completeness_filter(
    catalog: pd.DataFrame,
    completeness_table: pd.DataFrame,
) -> pd.DataFrame:
    """
    Args:
        catalog: the catalog to be filtered with columns
            - CSZ_ID
            - magnitude
            - time
        completeness_table: completeness table with columns
            - time
            - magnitude_completeness

    Returns:
        pd.DataFrame: the filtered catalog
    """
    catalog["time"] = catalog["time"].astype("datetime64[s]")
    completeness_table["time"] = completeness_table["time"].astype(
        str).apply(pd.Timestamp)
    df_out = pd.DataFrame()
    for zone_id, grp in completeness_table.groupby("CSZ_ID"):
        events = catalog[catalog["CSZ_ID"] == zone_id]
        to_remove = np.zeros(len(events), dtype=bool)
        table = grp.sort_values(by=["time", "magnitude_completeness"],
                                ascending=False)

        mc_first, tc_first = table.iloc[0][["magnitude_completeness", "time"]]
        to_remove |= (events["magnitude"] < mc_first) & (
            events["time"] >= tc_first)
        for mc, tc in zip(table["magnitude_completeness"][1:], table["time"]):
            to_remove |= (events["magnitude"] < mc) & (events["time"] < tc)

        tc_last = table.iloc[-1]["time"]
        to_remove |= (events["time"] < tc_last)

        df_out = pd.concat([df_out, events[~to_remove]])
    return df_out.sort_values(by="time")


def completeness_filter(
    catalog_path: os.PathLike,
    completeness_file: os.PathLike,
    csz_shapefile: os.PathLike | None = None,
    output_path: None | os.PathLike = None,
) -> pd.DataFrame:
    """
    Filter the catalog by a completeness table.

    A completeness table must contain at least the columns:
    CSZ_ID, Mc, Time
    For example with values:
    SZ01, 4.4, 1771
    SZ01, 5.2, 1613
    SZ01, 5.8, 1314
    ...
    SZ02, 3.8, 1852
    SZ02, 4.4, 1812
    ...

    time is a timestamp of the form "YYYY-MM-DD HH:MM:SS",
    at least the year must be provided.
    Mc is the magnitude of completeness.
    For example, (SZ01, 4.4, 1771) means that the catalog is complete
    for magnitudes >= 4.4  after 1771.

    Assume Time_i, Mc_i are the time and magnitude of completeness
    for the i-th row of the completeness table (descending by Time).

    Filtering removes events that
    - are assigned no zone
    - occured after Time_0 and have magnitude < Mc_0
    - occured before Time_n-1
    - occured between Time_i and Time_i+1 and have magnitude < Mc_i+1

    All events with magnitude < Mc that occured before Time are removed.

    Args:
        catalog_path: path to the catalog csv file, with columns
            - latitude
            - longitude
            - time
            - magnitude
        completeness_file: path to the completeness file, with columns
            - CSZ_ID
            - Mc
            - Time
        csz_shapefile: path to the completeness-superzone shapefile
        output_path: optional path to write the filtered catalog as a csv file

    Returns:
        pd.DataFrame: the filtered catalog
    """
    catalog = pd.read_csv(catalog_path)

    if csz_shapefile is not None:
        csz = gpd.read_file(csz_shapefile)
        catalog = add_zone_ids(catalog, csz, zone_column_name="CSZ_ID")

    completeness_table = pd.read_csv(completeness_file)

    # internally use lower case columns names
    completeness_table = completeness_table.rename(columns={
        "Time": "time",
        "Mc": "magnitude_completeness",
    })

    df_out = _completeness_filter(catalog, completeness_table)
    if output_path is not None:
        df_out.to_csv(output_path, index=False)
    return df_out
