import random
from typing import Literal

import pandas as pd


def replace_depths(catalog: pd.DataFrame,
                   default_depth: float,
                   boundary_km: float,
                   method: Literal['fixed', 'sampled'] = 'sampled'
                   ) -> pd.DataFrame:
    """
    Replaces the nan values in the 'depth' column of the catalog,
    based on which tectonic source zone the event is located in.

    Args:
        catalog: The catalog with required columns: 'depth'
            and 'TECTO_ID' if method is 'sampled'
        default_depth: The default depth to replace the nan values with
        boundary_km: The maximum depth to sample from
        method: The method to replace the nan values
            - fixed: replace the nan values with the default_depth
            - sampled: replace the nan values with random samples
                from the non-nan depths of the same tectonic source zone,
                having depths smaller than boundary_km. If no such depths
                are found, nan values are replaced with default_depth.

    Returns:
        the catalog with the nan 'depth' values replaced
    """
    if 'depth' not in catalog.columns:
        raise ValueError('column depth not found in the catalog')

    if method == 'fixed':
        catalog["depth"] = catalog["depth"].fillna(default_depth)
    elif method == 'sampled':
        if 'TECTO_ID' not in catalog.columns:
            raise ValueError('column TECTO_ID not found in the catalog')
        unassigned = catalog[catalog['TECTO_ID'].isna()]
        catalog.loc[unassigned.index,
                    'depth'] = unassigned['depth'].fillna(default_depth)

        for _, events in catalog.groupby('TECTO_ID'):
            catalog.loc[events.index,
                        'depth'] = _sample_replace_depths(events['depth'],
                                                          default_depth,
                                                          boundary_km)
    else:
        raise ValueError(f"Unknown method: {method}")
    return catalog


def _sample_replace_depths(depths: pd.Series, default_depth: float,
                           boundary_km: float) -> pd.Series:
    """
    Returns a copy of the input series with nan values replaced by random
    samples from the non-nan depths that are smaller than boundary_km.
    If no such depths are found, the nans are replaced by the default_depth.

    Args:
        depths: The series with nan values to be replaced
        default_depth: The default depth to replace the nan values with
        boundary_km: The maximum depth to sample from

    Returns:
        The series with the nan values replaced
    """
    depths = depths.copy()
    to_replace = depths[depths.isna()].index
    candidates = depths.loc[depths.notnull() & (depths < boundary_km)]
    if len(candidates) == 0:
        return depths.fillna(default_depth)

    new_depths = random.choices(candidates.values, k=len(to_replace))
    depths[depths.isna()] = new_depths
    return depths
