import os

import geopandas as gpd
import numpy as np
import pandas as pd
from numpy.testing import assert_almost_equal

from psha.common.associate_zones import add_zone_ids
from psha.harmonize.depth_replacement import replace_depths


def test_replace_depths():
    catalog = pd.DataFrame({'TECTO_ID': ["A", "A", "B", "B", "C"],
                            'depth': [10.2, np.nan, 20.0, np.nan, np.nan]})
    default_depth = 15
    boundary_km = 100
    catalog = replace_depths(catalog, default_depth,
                             boundary_km, 'fixed')
    expected_depths = np.array(
        [10.2, default_depth, 20.0, default_depth, default_depth])
    assert catalog['depth'].notnull().all()
    assert_almost_equal(catalog['depth'].values, expected_depths)


def test_replace_depths_sampled():
    catalog = pd.DataFrame({'TECTO_ID': ["A", "A", "B", "B", "C"],
                            'depth': [10.2, np.nan, 20.0, np.nan, np.nan]})
    default_depth = 15
    boundary_km = 100
    catalog = replace_depths(catalog, default_depth,
                             boundary_km, 'sampled')
    # Note: for tecto id 1 and 2, there is only one depth value
    # to sample from so it must be this
    # For the tecto id 3, we must take the default depth
    expected_depths = np.array([10.2, 10.2, 20.0, 20.0, default_depth])
    assert catalog['depth'].notnull().all()
    assert_almost_equal(catalog['depth'].values, expected_depths)


def test_replace_depths_boundary():
    catalog = pd.DataFrame({'TECTO_ID': ["A", "A", "A", "A", "A"],
                            'depth': [10.0, 10.0, 10.0, 40.0, np.nan]})
    default_depth = 15
    boundary_km = 30.0

    catalog = replace_depths(catalog, default_depth,
                             boundary_km, 'sampled')
    # only using depths strictly less than boundary_km
    expected_depths = np.array([10.0, 10.0, 10.0, 40.0, 10.0])
    assert catalog['depth'].notnull().all()
    assert_almost_equal(catalog['depth'].values, expected_depths)


def test_replace_depths_unassigned():
    catalog = pd.DataFrame({'TECTO_ID': ["A", "B", np.nan, "A", np.nan],
                            'depth': [10.0, np.nan, np.nan, 40.0, np.nan]})
    default_depth = 15
    boundary_km = 100

    catalog = replace_depths(catalog, default_depth,
                             boundary_km, 'fixed')
    # for events without tectionic id, we must use the default depth
    expected_depths = np.array(
        [10.0, default_depth, default_depth, 40.0, default_depth])
    assert catalog['depth'].notnull().all()
    assert_almost_equal(catalog['depth'].values, expected_depths)


def test_on_catalogue():
    shape_file = os.path.join(os.path.dirname(
        __file__), "data", "eshm20_tecto_winGT_fs017_dm02_results_v12e.shp")

    catalogue_path = os.path.join(os.path.dirname(
        __file__), "data", "depth_test.csv")

    df = pd.read_csv(catalogue_path)

    # add random nan to the catalogue depth
    df.loc[0:10, 'depth'] = np.nan
    df.loc[400:600, 'depth'] = np.nan

    df = add_zone_ids(df, gpd.read_file(shape_file),
                      zone_column_name='TECTO_ID')

    df_sampled = replace_depths(df, default_depth=10,
                                method='sampled', boundary_km=60.0)
    assert (df_sampled['depth'].notnull().all()
            ), "not all nan values were replaced"

    df_fixed = replace_depths(df, default_depth=10,
                              method='fixed', boundary_km=60.0)
    assert (df_fixed['depth'].notnull().all()
            ), "not all nan values were replaced"
