import numpy as np


def anderson_luco_1983_a_value(b_value: float,
                               area_fs: float,
                               max_mag: float,
                               slip_rate: float,
                               shear_modulus: float,
                               min_mag: float = 0.0,
                               ) -> float:
    """
    Compute a-values (intercept of the Gutenberg-Richter relationship)
    at min_mag for the Anderson and Luco (1983) N2 arbitrary area model.

    Args:
         b_value: b-value of the Gutenberg-Richter relationship
         area_fs: fault rupture area in km^2
         max_mag: maximum magnitude
         slip_rate: slip rate of the fault in mm/yr
         shear_modulus: shear modulus of the fault in GPa
         min_mag: minimum magnitude

    Returns:
       a_GR: Gutenberg-Richter a-value at Mw=min_mag
    """
    d_value = 1.5
    d_bar = d_value * np.log(10.0)
    b_bar = b_value * np.log(10.0)

    # Convert units
    # shear_modulus GPa   -> dyn/cm^2
    # area_fs       km^2  -> cm^2
    # slip_rate     mm/yr -> cm/yr
    slip_moment = (shear_modulus * 1e10) * \
        (area_fs * 1e10) * \
        (slip_rate / 10.0)
    delta_m = max_mag - min_mag
    moment_magnitude = 10.0 ** ((1.5 * float(max_mag)) + 16.05)
    a_2 = ((d_bar - b_bar) / b_bar) * \
        (slip_moment / moment_magnitude)
    occurence_rate = a_2 * (np.exp(b_bar * delta_m) - 1.0) * (delta_m > 0.0)
    # Number of events with M >= min_mag
    # N(M >= min_mag) = 10**a
    a_value = np.log10(occurence_rate)
    return round(a_value, 4)
