"""
Convert the raw catalogs into similar formats for further processing.
* Input:
  * `EMEC_final*.csv`  : European-Mediterranean Earthquake Catalog
  * `SERAcatalogue*`(`.csv` or `.xlsx`) : Seismology and Earthquake Engineering
                                          Research Infrastructure Alliance for
                                          Europe Catalog
  * `control_parameters.json` : details used to specify versions, file locations
                               & names, and column mapping
* Output:
  * `hmtk_cat_EMEC_Instrumental_*.csv` containing events on or after 1900 A.D.
  * `hmtk_cat_EMEC_Historical_*.csv` containing events prior to 1900 A.D.
  * `hmtk_cat_ESHM20_Historical_v1.0*.csv`

python reimplementation of : run_catalog_formatting_all.R

"""
import json
import os
import random

import geopandas as gpd
import numpy as np
import pandas as pd

from utils import munge_dates


def convert_sera(params) -> pd.DataFrame:
    """
    Make a version of the SERA catalog that is compatible with this workflow,
    saved as ESHM20 historical
    :return: sera historical catalog
    """
    # line 40 - 78 of run_catalog_formatting_all.R
    # Parameters

    filename = params['in_file']
    if filename.startswith('xls') or filename.endswith('xlsx'):
        sera: pd.DataFrame = pd.read_excel(pd.ExcelFile(filename), 'catalogue')
    else:
        sera = pd.read_csv(filename)

    # modify the SERA column names
    sera = sera.rename(index=str, columns=json.loads(params['oq_colmap']))

    # add columns that do not yet exist
    sera['Agency'] = params['agency']
    sera = sera.fillna(value={'SheecID': -1})
    sera = sera.dropna(subset=['magnitude'])
    sera = munge_dates(sera)

    sera = sera[json.loads(params['output_csv_cols'])]

    return sera


def convert_emec(params) -> pd.DataFrame:
    """
    Create a version of the EMEC catalog that is compatible with this workflow.
    Creates both historical and instrumental catalogs. But as of Jan2020, only
    the instrumental catalog is used.

    :return: sera instrumental catalog
    """
    # line 18 - 37 of run_catalog_formatting_all.R
    # different versions of the EMEC file use different separators.

    filename = params['in_file']
    emec = pd.read_csv(filename, sep=';')
    emec = emec.rename(index=str, columns={'eventID': 'SrcCatID'})  # eventID is used later. let's not confuse it
    emec = emec.set_index('SrcCatID', drop=True)
    emec = munge_dates(emec)
    emec['Agency'] = params['agency']

    # modify the column names
    emec = emec.rename(index=str, columns=json.loads(params['oq_colmap']))
    emec['uncertainty'] = np.nan
    emec.uncertainty = emec.uncertainty.astype(float)
    emec = emec[json.loads(params['output_csv_cols'])]

    return emec


def get_replacement_depths(depths: pd.Series, sample_method: str, default, boundary_km):  # ->pd.Series:
    """
    Replace nan values, either with a fixed value or by randomly sampled depths
    uses depths of shallow earthquakes only.
    """
    # replaces nan values
    depths = depths.copy()
    to_replace = depths[depths.isnull()].index
    candidates = depths.loc[depths.notnull() & (depths < boundary_km)]
    if (sample_method == 'fixed') or (len(candidates) == 0):
        if default is not None:
            depths = depths.fillna(default)
    elif sample_method == 'sampled':
        new_depths = random.choices(candidates.values, k=len(to_replace))
        depths.loc[to_replace] = new_depths
    return depths


def load_and_merge_catalogs(common_cols, historical, instrumental) -> pd.DataFrame:
    """

    Args:
        common_cols:
        historical:
        instrumental:

    Returns:

    """
    if isinstance(historical, pd.DataFrame):
        historical_catalog = historical
    else:
        historical_catalog = pd.read_csv(historical)

    # The instrumental catalog comes with a polygon field (string),
    # but the historical does not
    historical_catalog['polygon'] = ""

    if isinstance(instrumental, pd.DataFrame):
        instrumental_catalog = instrumental
        if ('SrcCatID' not in instrumental_catalog) and (instrumental_catalog.index.name == 'SrcCatID'):
            instrumental_catalog.index.rename('index', inplace=True)
            instrumental_catalog['SrcCatID'] = instrumental_catalog.index
    else:
        instrumental_catalog = pd.read_csv(instrumental)
    instrumental_catalog.loc[:, 'SrcCatID'] = instrumental_catalog['SrcCatID'].astype(str)
    joint_catalog = historical_catalog[common_cols].copy()

    joint_catalog = joint_catalog.append(instrumental_catalog[common_cols], ignore_index=True)
    joint_catalog.sort_values(by=['year', 'month', 'day', 'hour', 'minute', 'second'], inplace=True)
    joint_catalog = joint_catalog.reset_index()
    return joint_catalog


def _remove_duplicated_events(column_name, catalogue, save=False, logging_dir="."):
    """

    Args:
        column_name:
        catalogue:
        save:
        logging_dir:

    Returns:

    """
    dupidx = catalogue.index.duplicated(keep=False)
    if dupidx.any():
        these_dupes = {xx: list(catalogue.loc[xx, column_name]) for xx in catalogue.loc[dupidx].index.unique()}
        if save:
            catalogue.loc[dupidx].to_file(logging_dir + "/eq_duplicates_" + column_name + ".shp")
        catalogue = catalogue.loc[~catalogue.index.duplicated(keep="first")]
        catalogue.loc[catalogue[column_name].isna()].to_file(logging_dir + "/eq_unassigned_" + column_name + ".shp")
    else:
        these_dupes = {}
    return catalogue, these_dupes


def harmonize_catalogs(params, historical=None, instrumental=None):
    """

    Args:
        params:
        historical:
        instrumental:

    Returns:

    """
    from helpers.configuration import config

    if params['depth_replacement_method'] == "fixed":
        default_depth_shallow = params.getfloat('default_shallow_km')
        default_depth_deep = params.getfloat('default_deep_km')
    else:
        default_depth_shallow = None
        default_depth_deep = None

    if historical is None:
        historical = pd.read_csv(params['sera_file'])
    if any(historical['year'] > params.getint('split_year')):
        raise (
            ValueError(f"Historical catalog has events occurring after {params.split_year}. {max(historical['year'])}"))

    if instrumental is None:
        instrumental = pd.read_csv(params.emec_instrumental_file)
    if any(instrumental['year'] < params.getint('split_year')):
        raise (ValueError(f"Instrumental catalog has events prior to 1900. {min(instrumental['year'])}"))

    if 'SrcCatID' not in historical:
        historical['SrcCatID'] = ''
    if 'SrcCatID' not in instrumental:
        instrumental['SrcCatID'] = ''

    historical.loc[:, 'SrcCatID'] = historical['SrcCatID'].astype(str)

    if instrumental.index.name == "SrcCatID":
        instrumental.loc[:, 'SrcCatID'] = instrumental.index.astype(int).astype(str)
    else:
        instrumental.loc[:, 'SrcCatID'] = instrumental['SrcCatID'].astype(int).astype(str)
    print("type of incoming eventID>> hist:", historical['SrcCatID'].dtype, "inst:", instrumental['SrcCatID'].dtype)

    joint_catalog = load_and_merge_catalogs(json.loads(params['common_cols']), historical, instrumental)
    joint_catalog = joint_catalog.rename(index=str,
                                         columns={'Agency': 'source_catalog',
                                                  'SrcCatID': 'source_eventID'})
    joint_catalog['source_eventID'] = joint_catalog['source_eventID'].astype(str)
    joint_catalog.index = pd.Series(10 * 1_000_000 + joint_catalog.index.astype(int) + 1, name='eventID')
    joint_catalog['depth_orig'] = joint_catalog.depth

    tecto = gpd.read_file(config("primary")["tecto"]["shapefile"])
    tecto.index = tecto['TECTO_ID']
    tecto = tecto.sort_index()
    geojoint_catalog = gpd.GeoDataFrame(joint_catalog, geometry=gpd.points_from_xy(joint_catalog['longitude'],
                                                                                   joint_catalog['latitude']))
    geojoint_catalog.loc[:, 'TECTO_ID'] = None
    for tid, row in tecto.iterrows():
        geojoint_catalog.loc[geojoint_catalog.within(row.geometry), 'TECTO_ID'] = tid

    # replace invalid depths with either sampled data or the shallow_files default
    for tecto_id, events in joint_catalog.groupby('TECTO_ID'):
        depths_in_tecto = events['depth']
        replacement_method = 'fixed' if pd.isnull(tecto_id) else params['depth_replacement_method']
        joint_catalog.loc[events.index, 'depth'] = get_replacement_depths(depths_in_tecto,
                                                                          replacement_method,
                                                                          default_depth_shallow,
                                                                          params.getfloat('depth_boundary_km'))

    for tecto_id in joint_catalog["TECTO_ID"].unique():
        these_events = joint_catalog["TECTO_ID"] == tecto_id
        depths_in_tecto = joint_catalog.loc[these_events, 'depth']
        replacement_method = 'fixed' if pd.isnull(tecto_id) else params['depth_replacement_method']
        joint_catalog.loc[these_events, 'depth'] = get_replacement_depths(depths_in_tecto,
                                                                          replacement_method,
                                                                          default_depth_shallow,
                                                                          params.getfloat('depth_boundary_km'))

    precisions = {'latitude': 3, 'longitude': 3, 'depth': 3}
    for fld, prec in precisions.items():
        joint_catalog[fld] = pd.np.round(joint_catalog[fld], decimals=prec)

    joint_catalog = joint_catalog.drop(columns=['geometry'])
    joint_catalog.index.name = 'eventID'

    return joint_catalog


def create_all_catalogs(config_folder, dry_run=True):
    """
    Read original SERA and EMEC catalog, creating instru, hist, joint, shallow, deep cats

    Args:
        config_folder: location of config file
        dry_run:

    Returns:

    """
    from helpers.configuration import change_config_folder, config
    change_config_folder(config_folder)
    joint_params = config("catalog")["MasterCatalogCreation"]

    print('processing sera')
    sera_params = config("catalog")["SERA"]
    sera_cat = convert_sera(sera_params)

    print('processing emec')
    emec_params = config("catalog")["EMEC"]
    emec = convert_emec(emec_params)

    # split the catalog into historical and instrumental portions
    emec_historical = emec[emec.year < joint_params.getint('split_year')]
    emec_instrumental = emec[emec.year >= joint_params.getint('split_year')]

    print('combining & splitting by depth')
    joint_catalog = harmonize_catalogs(joint_params, historical=sera_cat, instrumental=emec_instrumental)

    shallow_cat = joint_catalog.loc[joint_catalog.depth < joint_params.getfloat("depth_boundary_km")]
    deep_cat = joint_catalog.loc[joint_catalog.depth >= joint_params.getfloat("depth_boundary_km")]

    if not dry_run:
        out_dir = joint_params['out_dir']
        os.makedirs(out_dir, exist_ok=True)
        sera_cat.to_csv(joint_params['sera_out'], index=False)
        emec_historical.to_csv(joint_params['emec_out_hist'])
        emec_instrumental.to_csv(joint_params['emec_out_inst'])
        joint_catalog.to_csv(joint_params["catalog_file_joint"], na_rep='')
        shallow_cat.to_csv(joint_params["catalog_file_shallow"], na_rep='')
        deep_cat.to_csv(joint_params["catalog_file_deep"], na_rep='')


if __name__ == "__main__":
    create_all_catalogs(config_folder='../working', dry_run=False)
