"""
Code used to decluster the catalog, based on configuration files

Decluster output files have:
   REASENBERG: columns for EventID, ClusterIndex,ClusterFlag
   WINDOWED: columns for EventID, CluserIndex, ClusterFlag
   ZALIAPIN: columns for EventID, IndependenceProbability.
"""
import json
import multiprocessing
import os
from typing import Dict

import geopandas as gpd
import pandas as pd
from openquake.hmtk.seismicity.declusterer.dec_gardner_knopoff import GardnerKnopoffType1
from openquake.hmtk.seismicity.declusterer.dec_reasenberg import Reasenberg
from openquake.hmtk.seismicity.declusterer.dec_zaliapin import Zaliapin
from openquake.hmtk.seismicity.declusterer.distance_time_windows import TIME_DISTANCE_WINDOW_FUNCTIONS

# clusterflag meanings
mainshock_flag = 0
aftershock_flag = 1
foreshock_flag = -1


def apply_declustering_to_catalog(config_folder):
    """decluster catalog according to parameters within the configuration file
    adding columns to original catalog


    Args:
        config_folder: location of configuration file used to guide declustering

    Returns:
        None

        Each declustering result will be saved individually.

        The final catalog will be saved with '_declust' appended to the end,
        and is saved as both comma-separated and shape files


    """
    from helpers.configuration import change_config_folder
    from helpers.configuration import config

    change_config_folder(config_folder)

    # get the catalogue(s) from file
    catalogue_file = config("catalog")["MasterCatalogCreation"]["catalog_file_shallow"]
    print(f'processing {catalogue_file}')
    # the catalog name becomes the base name for each file
    root_filename = os.path.split(catalogue_file)[1].rsplit('.', 1)[0]  # added on the fly
    catalogue = load_oq_catalogue(catalogue_file)

    generate_reasenberg()  # creates additional reasenberg decluster sections
    do_all_declustering(catalogue, root_filename)

    # add columns back to catalog

    # use pandas instead of openquake to sidestep catalog processing
    # that loses required columns/fields

    catalogue = pd.read_csv(catalogue_file, index_col='eventID', low_memory=False)
    catalogue = catalogue.drop(columns=[d for d in ['polygon', 'index'] if d in catalogue])

    results_folder = config("decluster")["General"]["declust_results_folder"]

    # add each set of decluster results to the catalog

    mapping = {}
    id = 0
    for filename in os.listdir(results_folder):
        if not filename.endswith('.csv'):
            continue
        id += 1
        decl_fullfile = os.path.join(results_folder, filename)
        add_decluster_results_to_catalog(catalogue, decl_fullfile)

        # get name from file
        with open(decl_fullfile, 'r') as file:
            l1 = file.readline()
            dectype = l1.split()[1]
            cn = l1.split('column=')[1].replace('"', '').strip()
            mapping[cn] = f"{dectype[:min(5, len(dectype))]}_n{id:03d}"

    with open(os.path.join(results_folder, 'shapefile_columns.json'), 'w') as file:
        print(json.dumps({v: k for k, v in mapping.items()}), file=file)

    print(catalogue.columns)
    catalogue.to_csv(catalogue_file.replace('.csv', '_declust.csv'))
    gpd.GeoDataFrame(catalogue,
                     geometry=gpd.points_from_xy(catalogue['longitude'],
                                                 catalogue['latitude'],
                                                 catalogue['depth'],
                                                 crs="epsg:4326")
                     ).rename(columns=mapping).to_file(
        catalogue_file.replace('.csv', '_declust.shp'))


def load_oq_catalogue(filename):
    """load earthquake catalog as openquake Catalogue, sorted chronologically

    Args:
        filename:

    Returns:
        openquake Catalogue
    """
    from openquake.hmtk.parsers.catalogue import CsvCatalogueParser

    print(multiprocessing.cpu_count())
    print(os.getcwd())
    print(filename)
    parser = CsvCatalogueParser(filename)
    catalogue = parser.read_file(start_year=1000, end_year=2015)
    print("\nNumber of events = %g" % catalogue.get_number_events())
    print("LonLat range       = %.4f<>%.4f and %.4f<>%.4f" % catalogue.get_bounding_box())
    catalogue.sort_catalogue_chronologically()
    print("Catalog sorting done!")
    return catalogue


def generate_reasenberg():
    """
    1. create a new section for each declustering combination, created from
       the [Generate reasenberg] section of the decluster ini.
    2. Update the ${General:declusters_to_append} with all new options

    Then,

    Returns:

    """
    import itertools
    from helpers.configuration import config

    # change_config_folder('../working')
    if "Generate reasenberg" not in config("decluster"):
        return

    cfg = config("decluster")["Generate reasenberg"]
    dpr = json.loads(cfg["declust_params"])
    # avoid duplication of permutaions:
    # because interaction_abbrev is coupled to interaction_formula
    # crate the formula->abbrev map, then remove abbrev from dict,
    # and then recalculated once permutations are done
    abbrmap = {k: v for k, v in zip(dpr["interaction_formula"], dpr["interaction_abbrev"])}
    del dpr['interaction_abbrev']

    # create dict of option permutations
    keys, values = zip(*dpr.items())
    params = [dict(zip(keys, v)) for v in itertools.product(*values)]

    # add interaction_abbrev back into permutations
    for p in params:
        p['interaction_abbrev'] = abbrmap[p['interaction_formula']]

    # now, add sections back into config file.
    dcfg = config("decluster")
    done = set()
    for n, p in enumerate(params):
        name = "rb" + '_'.join(f"{k.split('_')[-1]}{'_' if '_' in k else ''}{v}" for k, v in p.items())
        dcfg.add_section(name)
        dcfg[name]['method'] = 'reasenberg'
        dcfg[name]['name'] = name
        dcfg[name]['column'] = f"decl_rb{n:03d}"
        dcfg[name]['dm_correction'] = cfg["dm_correction"]
        dcfg[name]['declust_params'] = json.dumps(p)
        dcfg[name]['completeness_file'] = cfg.get('completeness_file', raw=True)
        dcfg['General']['declusters_to_append'] = ','.join([dcfg['General']['declusters_to_append'], name])
        print('\n', name, '\n', p)
        if name in done:
            raise ValueError("Duplication!")
        done.add(name)


def do_all_declustering(catalogue, root_filename: str):
    """perform each requested decluster

    Args:
        catalogue:
        root_filename:

    Returns:

    """
    from helpers.configuration import config

    do_methods = config("decluster")["General"]["declusters_to_append"].split(',')
    dry_run = config("decluster")["General"].getboolean("dry_run")

    interpretation_details = {}
    for n, method_section in enumerate(do_methods):
        print(f"Declustering: {n + 1} of {len(do_methods)}")
        method_deets = config("decluster")[method_section]
        mpars = json.loads(method_deets['declust_params'])
        print(f"{method_section} declustering")
        fn_args = (catalogue, root_filename, mpars, method_section, method_deets['column'])
        if method_deets["method"] == "windowed":
            saved_to = windowed_declustering(*fn_args, dry_run=dry_run)
        elif method_deets["method"] == "reasenberg":
            saved_to = reasenberg_declustering(*fn_args, dry_run=dry_run)
        elif method_deets["method"] == "zaliapin":
            saved_to = zaliapin_declustering(*fn_args, dry_run=dry_run)
        else:
            raise ValueError("unknown declustering method")
        interpretation_details[method_deets['column']] = {"opts": dict(method_deets),
                                                          "file": saved_to}
    with open(os.path.join(config("decluster")["General"]["declust_results_folder"],
                           'decluster_parameters_per_column.json'), "w") as file:
        print(json.dumps(interpretation_details), file=file)


def add_decluster_results_to_catalog(catalogue, declustresult_file):
    """add columns to catalog for each decluster result file in declust_results path

    Args:
        catalogue:
        declustresult_file:

    Returns:

    """
    from helpers.configuration import config

    # declustresult_file = os.path.join('../working/eq_cat/declust_results', filename)
    with open(declustresult_file, 'r') as file:
        l1 = file.readline().replace(',', '').strip().split(' ')
        dft = l1[1]
        name = [part for part in l1 if part.startswith('name=')][0].split('"')[1]
        colname = [part for part in l1 if part.startswith('column=')][0].split('"')[1]
    dr = pd.read_csv(declustresult_file, header=2, index_col='EventID')
    if dft.lower() == 'zaliapin':
        cutoffprob = config("decluster")[name].getint('cutoff_prob') / 100
        # catalogue[colname] = zaliapin_auto_assign_mainshocks(dr)
        catalogue[colname] = (dr['ProbIndep'] >= cutoffprob)
        catalogue.loc[:, colname] = catalogue[colname].astype(int)

    elif dft.lower() in ['reasenberg', 'windowed']:
        catalogue[colname] = (dr['ClusterFlag'] == mainshock_flag) | (dr['ClusterFlag'] == foreshock_flag)
        catalogue.loc[:, colname] = catalogue[colname].astype(int)
    else:
        raise ValueError('unknown declustring method')


def declusterfn(catalogue, declust_method, declust_config):
    total_events = catalogue.get_number_events()
    cluster_index, cluster_flag = declust_method.decluster(catalogue, declust_config)
    data = pd.DataFrame({'EventID': catalogue.data['eventID'],
                         'ClusterIndex': cluster_index,
                         'ClusterFlag': cluster_flag})
    total_mainshocks = sum(cluster_flag == 0)
    print("%g mainshocks out of %g Events (%.1f%%)" % (total_mainshocks, total_events,
                                                       total_mainshocks / total_events * 100))
    return data



def windowed_declustering(catalogue, root_filename, params, name, column=None, dry_run=False):
    """Decluster catalog using one of several pre-defined windows
    such as Gruenthal, GardinerKnopoff, or Uhrhammer
    
    Args:
        catalogue: oq-style Catalogue
        root_filename: filename, to be modified (no path)
        params: config section for this specific declustering attempt
        name:
        column:

    Returns:

    """
    from helpers.configuration import config

    declust_method = GardnerKnopoffType1()  # Create an `instance' of the tool
    dtwindow = TIME_DISTANCE_WINDOW_FUNCTIONS[params['window']]()
    # create generator for all configs.
    # win_abbr is only used for filename
    declust_config = {"time_distance_window": dtwindow,
                      "fs_time_prop": params['fs_time'],
                      "win_abbr": params['window']}

    # Run declustering
    if not dry_run:
        data = declusterfn(catalogue, declust_method, declust_config)
    # Save the catalog
    out_folder = config("decluster")["General"]["declust_results_folder"]
    outfile = os.path.join(out_folder,
                           f"{root_filename}" "_{win_abbr}_fs{fs_time_prop}.csv".format(**declust_config))
    if not dry_run:
        os.makedirs(out_folder, exist_ok=True)
        with open(outfile, 'w') as file:
            print(f'# Windowed name="{name}", column="{column}"', file=file)
            tdw = str(type(declust_config['time_distance_window']))
            print(f"# {json.dumps({**declust_config, 'time_distance_window': tdw})}", file=file)
            data.to_csv(file, index=False)
            print(f"----Saved to {outfile}")
    return outfile


def reasenberg_declustering(catalogue, root_filename: str, params: Dict, name, column=None, dry_run=False):
    """

    Args:
        catalogue: oq-style Catalogue
        root_filename: filename, to be modified (no path)
        params:  config section for this specific declustering attempt
        name:
        column:

    Returns:

    """
    from helpers.configuration import config

    reas_opts = params
    declust_method = Reasenberg()  # Create an `instance' of the tool
    errfact = reas_opts['errorfact']
    print(reas_opts)
    declust_config = dict(horiz_error=1.5 * errfact,
                          depth_error=2.0 * errfact,
                          **reas_opts)

    declust_config['xmeff'] = reas_opts['xmeff']
    if not dry_run:
        data = declusterfn(catalogue, declust_method, declust_config)

    # save the file

    abbr = reas_opts['interaction_abbrev']
    fnframe = "{outbase}_{abbr}_xmeff{xmeff}_rfact{rfact}_taumax{taumax}" \
              "_taumin{taumin}_errh{horiz_error}_errd{depth_error}_xk{xk}_p{P}.csv"
    fn = fnframe.format(outbase=root_filename, abbr=abbr, **declust_config)
    out_folder = config("decluster")["General"]["declust_results_folder"]
    outfile = os.path.join(out_folder, fn)
    if not dry_run:
        os.makedirs(out_folder, exist_ok=True)
        with open(outfile, 'w') as file:
            print(f'# Reasenberg name="{name}", column="{column}"', file=file)
            print(f"# {json.dumps(declust_config)}", file=file)
            data.to_csv(file, index=False)
            print("----Saved to %s" % outfile)
    return outfile


def zaliapin_declustering(catalogue, root_filename: str, params, name, column=None, dry_run=False):
    """

    Args:
        catalogue: oq-style Catalogue
        root_filename: filename, to be modified (no path)
        params:
        name:
        column:

    Returns:

    """
    from helpers.configuration import config

    declust_method = Zaliapin()  # Create an `instance' of the tool
    declust_config = {"fractal_dim": params['fdim'], "b_value": params['b'], "theta": params['theta']}

    total_events = catalogue.get_number_events()

    print(declust_config)

    if not dry_run:
        prob_indi = declust_method.decluster(catalogue, declust_config)
        data = pd.DataFrame({'EventID': catalogue.data['eventID'],
                             'ProbIndep': prob_indi})
        total_mainshocks = sum(prob_indi)
        print("%g mainshocks out of %g Events (%.1f%%)" %
              (total_mainshocks, total_events,
               total_mainshocks / total_events * 100))

    # Save the catalog

    out_folder = config("decluster")["General"]["declust_results_folder"]

    fn = "_Zaliapin_fdim{fractal_dim}_b{b_value}_theta{theta}.csv".format(**declust_config)
    outfile = os.path.join(out_folder, root_filename + fn)
    if not dry_run:
        os.makedirs(out_folder, exist_ok=True)
        with open(outfile, 'w') as file:
            print(f'# Zaliapin name="{name}", column="{column}"', file=file)
            print(f"# {json.dumps(declust_config)}", file=file)
            data.to_csv(file, index=False)
            print("----Saved to %s" % outfile)
    return outfile


def zaliapin_auto_assign_mainshocks(tb: pd.DataFrame):
    """return T/F series for mainshocks (with automatically determined cutoff)

    Args:
        tb: dataframe containing containing a column 'ProbIndep' from Zaliapin

    Returns:
        Series where the Independence Probability is greater than an
        automatically determined probability.


    The sum of the IndependenceProbability gives the approximate
    (expected) # of mainshocks, an conversely the T/F threshhold can be
    set by adjusting to get the expected number of mainshocks

    """

    n_main = round(sum(tb['ProbIndep']))
    print('n Main Events', n_main)
    tb: pd.DataFrame = tb.sort_values(by=['ProbIndep'], ascending=False)
    cutoffprob = tb.iloc[n_main]['ProbIndep']
    print('cutoff probability:', cutoffprob)
    return tb['ProbIndep'] >= cutoffprob


if __name__ == "__main__":
    apply_declustering_to_catalog('../working')
