import os
import tomllib
import warnings
from pathlib import Path

import pandas as pd
from seismostats.analysis.declustering import GardnerKnopoffType1
from seismostats.analysis.declustering.base import Declusterer
from seismostats.analysis.declustering.distance_time_windows import (
    GardnerKnopoffWindow, GruenthalWindow)


def decluster_mainshocks(input_path: os.PathLike,
                         declusterer: Declusterer,
                         output_column_name: str = "mainshock_flag",
                         output_path: os.PathLike | None = None,
                         verbose: bool = False) -> pd.DataFrame:
    """
    Declustering is applied to the input catalogue.
    The output catalogue contains the same columns as the input catalogue plus
    the columns with the mainshock flags for each declusterer.

    Args:
        input_path: path to the input catalogue as a csv file
        declusterer: a seismostats Declusterer to apply
        output_column_name: names the column for the mainshock flags
        output_path: optional path to write the output catalogue as a csv file
        verbose: if True, print additional information

    Returns:
        pd.DataFrame: the declustered catalogue
    """
    input_path = Path(input_path).resolve()
    output_path = Path(output_path).resolve()
    # Sanity checks
    if not input_path.exists():
        raise RuntimeError("The input catalogue does not exist.")
    if output_path.exists():
        warnings.warn("Output catalogue already exists.")
    if input_path == output_path:
        raise RuntimeError(
            "The input and output files are the same. "
            "This would overwrite the original file"
        )
    df = pd.read_csv(input_path)

    mainshocks = declusterer(df)

    df[output_column_name] = mainshocks
    if verbose:
        n_main = mainshocks.sum()
        n_total = mainshocks.size
        percents = n_main / n_total * 100
        print(f"{n_main} Mainshocks out of {n_total} Events \
            ({percents:02.1f}%)")
    if output_path is not None:
        df.to_csv(output_path, index=False)
    if verbose:
        print(f"Declustered catalogue written to {output_path}")
    return df


def main():
    with open("A01_config.toml", "rb") as f:
        config = tomllib.load(f)
    verbose = config.get("verbose")
    if verbose:
        print(f"config: {config}")
    input_catalog_path = config.get("input_catalog_path")
    output_column_name = config.get("output_column_name")
    output_catalog_path = config.get("output_catalog_path")
    input_path = Path(input_catalog_path).resolve()
    output_path = Path(output_catalog_path).resolve()

    declusterer_name = config.get("declusterer")
    declusterer_args = config["declusterer_args"]
    if declusterer_name == "GardnerKnopoffType1":
        window_type = declusterer_args.pop("time_distance_window")
        if window_type == "GardnerKnopoffWindow":
            declusterer_args["time_distance_window"] = GardnerKnopoffWindow()
        elif window_type == "GruenthalWindow":
            declusterer_args["time_distance_window"] = GruenthalWindow()
        else:
            raise ValueError(f"Unknown window type: {window_type}")
        declusterer = GardnerKnopoffType1(**declusterer_args)
    else:
        raise ValueError(f"Unknown declusterer: {declusterer_name}")

    decluster_mainshocks(input_path=input_path,
                         declusterer=declusterer,
                         output_column_name=output_column_name,
                         output_path=output_path, verbose=verbose)


if __name__ == "__main__":
    main()
