# Workflows A2

Workflow A21 prepares the input data for workflow A22, adding metadata to faults, based on which tectonic zone they are associated with.

Workflow A22 takes as input fault data, computes a-values using a statistical method for a combination of parameters, and saves Magnitude-Frequency Distribution (MFD) parameters as XML files for Openquake.


## Workflow A21: Assign TECTO metadata to Faults

1. Load Shapefiles
    - tectonic zone polygons (`tecto_shapefile`)
    - active fault lines (`faults_shapefile`)
2. Join faults with polygons
    - ensure that the tectonic zone and fault lines shae the same coordinate reference system (`projected_crs`)    
    - associate fault lines with polygons based on spatial containment of the centroid of the fault line
3. Extract and Merge Attributes
    - add relevant columns of the tectonic data to the associated faults (`column_map`)
    - e.g. for each fault `tct_b_val` is added as the `b_val` of the associated tectonic zone
4. Save Results
    - as geojson (`output_geojson`), can be used by A22
    - as a shapefile (`output_shapefile`), so far not further used
    - in crs `output_crs`

### Config
Sample `A21_config.toml`:
``` toml
tecto_shapefile = "test/data/eshm20_input_c_tecto_zones/eshm20_tecto_winGT_fs017_dm02_results_v12e.shp"
faults_shapefile = "test/data/eshm20_input_e_fs_model/eshm20_fs_model_v09e.shp"

output_shapefile = "output/shapefiles/eshm20_tecto_with_attributes.shp"
output_geojson = "output/shapefiles/eshm20_tecto_with_attributes.geojson"

projected_crs = "EPSG:32633"
output_crs = "EPSG:4326"

[column_map]
TECTO_ID = "TECTO_ID"
MAXMAG01 = "tct_MAXMAG01"
MAXMAG02 = "tct_MAXMAG02"
MAXMAG03 = "tct_MAXMAG03"
a_val = "tct_a_val"
b_val = "tct_b_val"
```

### Running
Simply run
``` bash
cd workflows_a/A2
python workflow_A21_assign_tecto_metadata_to_faults.py
```

This workflow was tested only manually, there are no automated tests.


### Faults within Tectos
Tecto and fault data visualized:
![](./assets/faults_within_tecto.png)
As an example for the association, two tectonic zones and the corresponding faults.
![](./assets/TSZ104.png)
![](./assets/TSZ220.png)


## Workflow A22: Compute Fault Statistics
This workflow computes GR a-values for faults
using a different combination of parameters
and writes nrml XML files with MFD parameters for further processing.

- Input
    - Reads fault data from the geojson format (`input_fault_db`)
- For each fault
    - properties of the fault are extracted
    - for maximum magnitudes (M) and slip rates (SR),
      low (L), average (A), and upper (U) values are computed
    - this defines a 3x3 parameter grid:
        |--------|--------|--------|
        | SRL_ML | SRL_MA | SRL_MU |
        | SRA_ML | SRA_MA | SRA_MU |
        | SRU_ML | SRU_MA | SRU_MU |


    - for each combiantion of the parameters,
       Gutenberg-Richter a-values are calculated (at `gr_minimum_magnitude`)

- Output
    - Writes nrml XML files for each paramter combination. Example output files, the name based on `model_name`:
    ``` text
    fs_model_eshm20_ver09e_aGR_SRA_MA_fMthrStatic.xml
    fs_model_eshm20_ver09e_aGR_SRA_ML_fMthrStatic.xml
    fs_model_eshm20_ver09e_aGR_SRA_MU_fMthrStatic.xml
    fs_model_eshm20_ver09e_aGR_SRL_MA_fMthrStatic.xml
    fs_model_eshm20_ver09e_aGR_SRL_ML_fMthrStatic.xml
    fs_model_eshm20_ver09e_aGR_SRL_MU_fMthrStatic.xml
    fs_model_eshm20_ver09e_aGR_SRU_MA_fMthrStatic.xml
    fs_model_eshm20_ver09e_aGR_SRU_ML_fMthrStatic.xml
    fs_model_eshm20_ver09e_aGR_SRU_MU_fMthrStatic.xml
    ```
    - Pickles the fault data for debugging (`output_pickle_dir`)


Please note that there is a static (fMthrVar) and variable (fMthrStatic) version:
if `magnitude_threshold = 5.9` is set then this will be used as a static magnitude threshold for all faults.
Otherwise, the fault magnitude is read from the attributes of each fault.


### Config
Sample `A22_config.toml`:
``` toml
input_fault_db = "test/data/eshm20_input_e_fs_model/fs_model_eshm20_ver09e.geojson"
model_name = "fs_model_eshm20_ver09e"

output_pickle_dir = "output/pickles/"
output_xml_dir = "output/nrmls/"

# if not set reads CLUSTER from the fault attrs
tectonic_reg = "Shallow Default"
# if not set reads FSLNAMmagnitude_threshold = 5.9
magnitude_threshold = 5.9

[anderson_luco]
# Shear modulus in GPa
shear_modulus = 30.0
# Gutenberg-Richter minimum magnitude
# WARNING: to reproduce previous results,
# using the hmtk anderson luco implementation
# set this to -bin_width/2
gr_minimum_magnitude = -0.05
```
With the `anderson_luco` option, we may configure the method used to calculate the a-values.

### Running
Running it
``` bash
cd workflows_a/A2
python workflow_A22_process_crustal_faults.py  
# test if the same xml files are produced
./test/diff_regression.sh   
```
`diff_regression.sh` compares the generated xml files.
This is a rudimentary check, e.g. it breaks if the formatting of xml changes although it still contains the same semantic data.
The regression data contains files for variable and static magnitude thresholds.
Hence, for the first time run this once with a static `magnitude_threshold = 5.9`, and once with this option commentec out to generate all of the output data.


## References
See also: [Google Doc: A2 Workflow Summary](https://docs.google.com/document/d/1iehAANydF_IQ1fR_WKZ-IH_5cB0cyzcahMzA95TGXA4/edit?usp=sharing)