# %%
import tomllib
from pathlib import Path

import geopandas as gpd
import matplotlib.pyplot as plt

# %%
"""
Workflow A21: Assign Tectonic Polygon Metadata to Faults

Associates metadata from tectonic polygons to crustal faults based on spatial
containment of the centroid of the fault lines within the polygons.
The output is to be used for Workflow A22.

Inputs:
- A shapefile containing the tectonic polygons
- A shapefile containing the crustal faults as lines

Configuration:
- A21_config.toml
- column_map: mapping of the columns from the tectonic polygons
    to be added to the fault lines.
- (optional) the CRS to save the output
- (optional) the CRS to reproject the input data


Outputs:
- A shapefile containing the faults with the tectonic polygon metadata
- A GeoJSON file containing the same data

Steps:
1. Load the tectonic polygons and the fault lines
2. Check the CRS of the data and reproject if needed
3. Spatially join the fault lines with the tectonic polygons
4. Save the output files
5. Optionally visualize the associations between the faults and the polygons

"""
# %%
with open("A21_config.toml", "rb") as f:
    config = tomllib.load(f)

# Shapefile containing Tectonic Polygons
tecto_shapefile_path = config.get("tecto_shapefile")
# Shapefile containing Crustal Faults as Lines
faults_shapefile_path = config.get("faults_shapefile")

output_shapefile_path = config.get("output_shapefile")
output_geojson_path = config.get("output_geojson")

tecto_shapefile_path = Path(tecto_shapefile_path).resolve()
faults_shapefile_path = Path(faults_shapefile_path).resolve()
output_shapefile_path = Path(output_shapefile_path).resolve()
output_geojson_path = Path(output_geojson_path).resolve()


PROJECTED_CRS = config.get("projected_crs", "EPSG:32633")
OUTPUT_CRS = config.get("output_crs", "EPSG:4326")

# attributes added to the faults from the zone they are joined with
# and renamed based on the following mapping
COLUMN_MAP = config["column_map"]

# %%
polygons_gdf = gpd.read_file(tecto_shapefile_path)
lines_gdf = gpd.read_file(faults_shapefile_path)
print(polygons_gdf.describe())
print(lines_gdf.describe())

assert set(COLUMN_MAP.keys()) <= set(polygons_gdf.columns), \
    "Columns missing in the tectonic dataframe"

# Clean columns from the faults if they are already present
remapped = list(set(COLUMN_MAP.values()) & set(lines_gdf.columns))
lines_gdf = lines_gdf.drop(columns=remapped)

# %%
# Check the current CRS and reproject if needed
print("Polygons original CRS:", polygons_gdf.crs)
print("Lines original CRS:", lines_gdf.crs)

# Reproject both to a suitable projected CRS if they're in a geographic CRS
if polygons_gdf.crs.is_geographic:
    polygons_gdf = polygons_gdf.to_crs(PROJECTED_CRS)
if lines_gdf.crs.is_geographic:
    lines_gdf = lines_gdf.to_crs(PROJECTED_CRS)

print("Polygons projected CRS:", polygons_gdf.crs)
print("Lines projected CRS:", lines_gdf.crs)

# %%
# Calculate centroids of the lines
lines_gdf['centroid'] = lines_gdf.geometry.centroid

# Ensure both geodataframes have the same CRS
if polygons_gdf.crs != lines_gdf.crs:
    lines_gdf = lines_gdf.to_crs(polygons_gdf.crs)

# Spatial join based on centroid containment within the polygons
joined_gdf = gpd.sjoin(lines_gdf.set_geometry('centroid'),
                       polygons_gdf,
                       how='left',
                       predicate='within')

print("Joined GeoDataFrame columns:", joined_gdf.columns)

# %%
print("column mapping:", COLUMN_MAP)
joined_gdf = joined_gdf.rename(columns=COLUMN_MAP)
result_gdf = lines_gdf.join(joined_gdf[list(COLUMN_MAP.values())])

# %%
# Step 4: Save result
output_shapefile_path.parent.mkdir(exist_ok=True, parents=True)
output_gdf = result_gdf.drop(columns='centroid').to_crs(OUTPUT_CRS)
output_gdf.to_file(output_shapefile_path)
print(f"wrote: {output_shapefile_path}")
output_geojson_path.parent.mkdir(exist_ok=True, parents=True)
output_gdf.to_file(output_geojson_path, driver="GeoJSON")
print(f"wrote: {output_geojson_path}")


# %%


def plot_fault_lines_within_polygons(polygons: gpd.GeoDataFrame,
                                     faults: gpd.GeoDataFrame) -> None:
    """
    Visualize all fault lines with centroids and the polygons they are within
    """
    _, ax = plt.subplots(figsize=(10, 10))
    polygons.boundary.plot(ax=ax, color='blue', label='Polygons')
    faults.geometry.plot(ax=ax, color='green', label='Lines')
    faults.set_geometry('centroid').plot(
        ax=ax, color='red', label='Centroids', markersize=5)

    plt.legend()
    plt.title("faults within the TECTO polygons")
    plt.show()


# plot_fault_lines_within_polygons(polygons_gdf, result_gdf)
# %%


def plot_single_tecto(polygons: gpd.GeoDataFrame,
                      faults: gpd.GeoDataFrame) -> None:
    """
    Create a plot for each tectionic polygon with the fault lines within it.
    """
    tecto_ids = polygons["TECTO_ID"].unique()
    for tecto_id in tecto_ids:
        fault = faults[faults["TECTO_ID"] == tecto_id]
        polygon = polygons_gdf[polygons_gdf["TECTO_ID"] == tecto_id]
        plot_fault_lines_within_polygons(polygon, fault)


# warning: this will create a lot of plots
# plot_single_tecto(polygons_gdf, result_gdf)
