# %%
import json
import pickle
import tomllib
from itertools import product
from pathlib import Path
from typing import Dict, List, Tuple, TypedDict

import numpy as np

from psha.mfd.anderson_luco import anderson_luco_1983_a_value

"""
Abbreviations
- GR :: Gutenberg-Richter
- MFD :: Magnitude Frequence Distribution
- MSR :: Magnitude-area Scaling Relationship
- SR :: Slip-Rate
- Mthr :: Magnitude threshold
- NRML :: Natural hazard's Risk Markup Language

This workflow computes GR a-values for faults
using a different combination of parameters
and writes nrml XML files with MFD parameters for further processing.

- Reads fault data from the geojson format
- For each fault
    - properties of the fault are extracted
    - for maximum magnitudes (M) and slip rates (SR),
      low (L), average (A), and upper (U) values are computed
    - this defines a 3x3 parameter grid
    - for each combiantion of the parameters,
       GR a-values are calculated (at gr_minimum_magnitude)
- Pickles the fault data (for debugging)
- Writes nrml XML files for each paramter combination

Set configuration options in A22_config.toml

The output files are named based on the model_name

if magnitude_threshold is provided it is used as
    minimum magnitude for the MFD of each fault
Othwerise,
    this is read for each fault from the attribute "FaultMthr"
"""

# %%
# WARNING: the order of the loops matters, must be the same
# as the parameters are generated in combi
PARAM_GRID_NAMES = [f"{sr}_{mm}"
                    for mm in ['ML', 'MA', 'MU']
                    for sr in ['SRL', 'SRA', 'SRU']]
# %%

with open("A22_config.toml", "rb") as f:
    config = tomllib.load(f)

output_xml_dir = Path(config.get('output_xml_dir')).resolve()
output_pickle_dir = Path(config.get('output_pickle_dir')).resolve()


# E.g. "Iceland Atlantic Active Region"
TECTONIC_REG = config.get('tectonic_reg')
SCALING_RELATION = config.get('scaling_relation')

# %%
default_params = config["anderson_luco"]
SHEAR_MODULUS = default_params.get("shear_modulus", 30.0)
GR_MIN_MAG = default_params.get("gr_minimum_magnitude", 0.0)


MAGNITUDE_THRESHOLD = config.get('magnitude_threshold')
USE_STATIC_MTHR = MAGNITUDE_THRESHOLD is not None
if USE_STATIC_MTHR:
    model_suffix = "fMthrStatic"
else:
    model_suffix = "fMthrVar"
model_name = config.get('model_name')


# %%


class FaultProperties(TypedDict):
    source_name: str
    fault_id: str
    fault_name: str
    scaling_relation: str
    aspect_ratio: float
    fault_lonlat: List[float]
    min_depth: float
    max_depth: float
    dip_avg: float
    rake_avg: float
    min_mag: float
    tectonic_region: str
    b_value: float
    area_avg: float
    max_mags: Tuple[float, float, float]
    slip_rates: Tuple[float, float, float]


def _preprocess_fault(fault) -> FaultProperties:
    """
    Extracts the attributes from the fault that are needed later.
    Applies normalization/sanitzation to some attributes.
    """
    fault_properties = fault["properties"]
    source_name = str(fault_properties["IDFS"]).replace(
        ".", "").replace("-", "")
    fault_name = str(fault_properties["FSNAME"]).replace(
        ".", "").replace("-", "")

    if SCALING_RELATION:
        scaling_relation = SCALING_RELATION
    else:
        scaling_relation = fault_properties["FSLNAME"]

    aspect_ratio = np.round(fault_properties["RUPASP"], 2)

    fault_lonlat = np.array(fault["geometry"]["coordinates"])

    # depth range
    min_depth = np.abs(fault_properties["MINDEPTH"])
    max_depth = np.abs(fault_properties["MAXDEPTH"])

    # dip range
    dip_avg = fault_properties["DIPAVG"]
    if (dip_avg > 90):
        dip_avg = dip_avg - 90.0

    rake_avg = fault_properties["RAKEAVG"]
    if (rake_avg > 180):
        rake_avg = rake_avg - 360.0

    b_value = fault_properties["tct_b_val"]

    if USE_STATIC_MTHR:
        min_mag = MAGNITUDE_THRESHOLD
    else:
        min_mag = fault_properties["FaultMthr"]

    if TECTONIC_REG:
        tectonic_region = TECTONIC_REG
    else:
        tectonic_region = fault_properties["CLUSTER"]

    # Slip rate (fault displacement rate)
    # Penalize faults with a large range between min and max slip rates
    slip_min = fault_properties["SRMIN"]
    slip_avg = fault_properties["SRAMEAN"]
    slip_max = fault_properties["SRMAX"]
    if np.round((slip_max / slip_min), 2) > 10.01:
        # Large variability in slip rates
        slip_avg = fault_properties["SRGMEAN"]

    # Fault area (average rupture area)
    area_avg = fault_properties["AREAAVG"]

    # Maximum magnitudes
    max_mag_min = np.round(fault_properties["fMAXMAG01"], 1)
    max_mag_mid = np.round(fault_properties["fMAXMAG02"], 1)
    # max_mag01_tct = np.round(fault_properties["tct_MAXMAG01"], 1)
    max_mag02_tct = np.round(fault_properties["tct_MAXMAG02"], 1)
    max_mag03_tct = np.round(fault_properties["tct_MAXMAG03"], 1)

    # Weighted computation of maximum magnitude
    if max_mag_mid <= 6.9:
        max_mag_max = max_mag_mid * 0.7 + max_mag02_tct * 0.3
    else:
        max_mag_max = max_mag_mid * 0.3 + max_mag03_tct * 0.7

    # Ensure mid magnitude does not exceed computed max magnitude
    if max_mag_mid > max_mag_max:
        max_mag_max = max_mag_mid + 0.6

    # Collect parameter combinations
    # We compute a_values for each pair of max_mags and slip_rates
    max_mags = [max_mag_min, max_mag_mid, max_mag_max]
    slip_rates = [slip_min, slip_avg, slip_max]

    return {
        "source_name": source_name,
        "fault_id": fault_properties["IDFS"],
        "fault_name": fault_name,
        "scaling_relation": scaling_relation,
        "aspect_ratio": aspect_ratio,
        "fault_lonlat": fault_lonlat,
        "min_depth": min_depth,
        "max_depth": max_depth,
        "dip_avg": dip_avg,
        "rake_avg": rake_avg,
        "min_mag": min_mag,
        "tectonic_region": tectonic_region,
        "b_value": b_value,
        "area_avg": area_avg,
        "max_mags": max_mags,
        "slip_rates": slip_rates,
    }

# %%


FaultId = str
FaultData = Dict[FaultId, FaultProperties]


def load_fault_data(input_path: Path) -> FaultData:
    """
    Loads fault data from the GEOJSON format

    Args:
        input_path: path to the geojson file

    Returns:
        fault_data: dict containing all relevant fault attributes
    """
    with open(input_path) as f:
        _data = json.load(f)
        res = {
            feature["properties"]["IDFS"]: _preprocess_fault(feature)
            for feature in _data["features"]
        }
    print("read ", input_path)
    print(f"total no. of faults: {len(res)}")
    return res


input_path = Path(config.get('input_fault_db')).resolve()
fault_data = load_fault_data(input_path)

# %%


def a_values_for_faults(fault_data: FaultData) -> Tuple[dict, dict]:
    """
    Computes a-values for the each fault using an external method,
    for each combination in the parameter grid.

    Teh results in a_values are mappings of the form:
    param_label -> fault_id -> a_value
    param_label -> fault_id -> fault_max_mag

    For example:
    a_values["SRA_ML"]["PTCF004"] = 3.45
    fault_max_mags["SRA_ML"]["PTCF004"] = 7.2

    Returns:
        a_values
        fault_max_mags
    """
    a_values = {label: dict() for label in PARAM_GRID_NAMES}
    fault_max_mags = {label: dict() for label in PARAM_GRID_NAMES}

    for fault_id, fault in fault_data.items():
        max_mags = fault["max_mags"]
        slip_rates = fault["slip_rates"]

        b_value = np.round(fault["b_value"], 3)
        area_avg = fault["area_avg"]

        # WARNING: the iteration order matters, the same as PARAM_GRID_NAMES
        combi = tuple(product(max_mags, slip_rates))

        # Compute a-values (intercept of magnitude-frequency relationship)
        for label, (max_mag, sr) in zip(PARAM_GRID_NAMES, combi):
            a_value = anderson_luco_1983_a_value(
                b_value=b_value,
                area_fs=area_avg,
                min_mag=GR_MIN_MAG,
                max_mag=max_mag,
                slip_rate=sr,
                shear_modulus=SHEAR_MODULUS
            )
            a_values[label][fault_id] = a_value
            fault_max_mags[label][fault_id] = np.round(max_mag, 3)

    print('computed a-values')
    return a_values, fault_max_mags


aGR, faultMaxMags = a_values_for_faults(fault_data)
# %%


def write_pickle(output_pickle: Path, obj: object):
    output_pickle.parent.mkdir(exist_ok=True, parents=True)
    with open(output_pickle, "wb") as f:
        pickle.dump(obj, f)
    print('pickled ', output_pickle)


# Pickle the fault_data for debugging
output_pickle = output_pickle_dir / f"{model_name}_{model_suffix}.pickle"
write_pickle(output_pickle, fault_data)


# %%
def nrml_xml(grid_label: str, fault_data) -> str:
    """
    Args:
        grid_label: to retrieve the computed a values and max_mags.
        fault_data

    Returns:
        xml_data: nrml xml file with simpleFaultSource elements
    """
    header = '<?xml version="1.0" encoding="utf-8"?>\n'
    header += '<nrml xmlns:gml="http://www.opengis.net/gml"\n'
    header += '      xmlns="http://openquake.org/xmlns/nrml/0.4">\n\n'

    newxml = header + '    <sourceModel name="' + \
        f"{model_name}_{model_suffix}" + '">\n\n'

    for fault_id, f_data in fault_data.items():
        source_name = f_data["source_name"]
        fault_name = f_data["fault_name"]
        scaling_relation = f_data["scaling_relation"]
        aspect_ratio = f_data["aspect_ratio"]
        fault_lonlat = f_data["fault_lonlat"]
        min_depth = f_data["min_depth"]
        max_depth = f_data["max_depth"]
        dip_avg = f_data["dip_avg"]
        rake_avg = f_data["rake_avg"]
        min_mag = f_data["min_mag"]
        tectonic_region = f_data["tectonic_region"]

        aValue = aGR[grid_label][fault_id]
        bValue = f_data["b_value"]
        maxMag = faultMaxMags[grid_label][fault_id]

        newxml += '        <simpleFaultSource id="' + source_name + \
            '" name="' + fault_name + \
            '" tectonicRegion="' + tectonic_region + '">\n'
        newxml += '            <simpleFaultGeometry>\n'
        newxml += '                <gml:LineString>\n'
        newxml += '                    <gml:posList>\n'
        for xy_string in fault_lonlat:
            xy_string = "".join(
                [str(xy_string).replace("[", "").replace("]", "")])
            newxml += '           ' + xy_string.ljust(10) + '\n'

        newxml += '                    </gml:posList>\n'
        newxml += '                </gml:LineString>\n'
        newxml += '                <dip>' + str(dip_avg) + '</dip>\n'
        newxml += '                <upperSeismoDepth>' + \
            str(min_depth) + '</upperSeismoDepth>\n'
        newxml += '                <lowerSeismoDepth>' + \
            str(max_depth) + '</lowerSeismoDepth>\n'
        newxml += '            </simpleFaultGeometry>\n'
        newxml += '            <magScaleRel>' \
            + scaling_relation + '</magScaleRel>\n'
        newxml += '            <ruptAspectRatio>' + \
            str(aspect_ratio) + '</ruptAspectRatio>\n'

        #  TODO(old) the incremental from the MBTK () to be done
        newxml += '          <truncGutenbergRichterMFD aValue="' \
            + str(aValue) + '"   bValue="' + str(round(float(bValue), 3)) + \
            '"  minMag="' + str(round(min_mag, 1)) + \
            '"  maxMag="' + str(round(maxMag, 1)) + '"/>\n'

        newxml += '            <rake>' + str(rake_avg) + '</rake>\n'
        newxml += '        </simpleFaultSource>\n\n'

    newxml += '    </sourceModel>\n'
    newxml += '</nrml>'

    return newxml

# %%


def write_xml(dir: Path):
    """
    Writes a nrml file for each parameter combination.

    E.g.
    fs_model_eshm20_ver09e_aGR_SRL_ML_fMthrVar.xml
    fs_model_eshm20_ver09e_aGR_SRA_ML_fMthrVar.xml
    ...

    Args:
        dir: directory to put the output files
    """
    output_xml_dir.mkdir(exist_ok=True, parents=True)
    for label in PARAM_GRID_NAMES:
        output_file = dir / f'{model_name}_aGR_{label}_{model_suffix}.xml'

        xml = nrml_xml(label, fault_data)

        with open(output_file, "w") as f:
            f.write(xml)

        print('wrote ', output_file)


write_xml(output_xml_dir)
